package de.unibw.i2.mi.oo;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Class representing classes.
 */
public class MIClass extends MIType {
	private final boolean isAbstract;
	private MIClass superClass;
	private final List<MIInterface> interfaces;
	private Map<MIMethodSignature, MIMethod> impl = new HashMap<MIMethodSignature, MIMethod>();

	public MIClass(String name, boolean isAbstract, MIClass superClass,
			MIInterface... interfaces) {
		this(name, isAbstract, superClass, Arrays.asList(interfaces));
	}

	public MIClass(String name, boolean isAbstract, MIClass superClass,
			List<MIInterface> interfaces) {
		super(name);
		this.superClass = superClass;
		this.interfaces = new ArrayList<MIInterface>(interfaces);
		this.isAbstract = isAbstract;
		if (isAbstract)
			addMethod(MIMethodSignature.UPCAST);
		else {
			MIMethod upcast = new MIMethod(this, MIMethodSignature.UPCAST);
			addMethod(upcast);
		}
	}

	public MIClass getSuperClass() {
		return superClass;
	}

	public boolean isAbstract() {
		return isAbstract;
	}

	public List<MIInterface> getInterfaces() {
		return interfaces;
	}

	@Override
	public List<? extends MIType> getDirectSupertypes() {
		List<MIType> supertypes = new ArrayList<MIType>(
				getInterfaces().size() + 1);
		if (getSuperClass() != null)
			supertypes.add(getSuperClass());
		supertypes.addAll(getInterfaces());
		return supertypes;
	}

	/**
	 * Adds a non-abstract method to this class. This method actually calls
	 * {@link MIType#addMethod(MIMethodSignature)} and links the given method as
	 * an implementation to the corresponding signature. Use this method for
	 * adding a non-abstract method to this type.
	 */
	public void addMethod(MIMethod method) {
		if (method.getImplementingClass() != this)
			throw new IllegalStateException("Wrong class");
		super.addMethod(method.getSig());
		impl.put(method.getSig(), method);
	}

	@Override
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("MIClass [name=").append(getName());
		if (getSuperClass() != null)
			buf.append(", superClass=").append(getSuperClass().getName());
		if (!getInterfaces().isEmpty()) {
			buf.append(", interfaces=[");
			boolean first = true;
			for (MIInterface si : getInterfaces()) {
				if (!first)
					buf.append(", ");
				first = false;
				buf.append(si.getName());
			}
			buf.append("]");
		}
		buf.append("]");
		return buf.toString();
	}

	private Path getPath() {
		Path path;
		if (getSuperClass() == null)
			path = new Path();
		else
			path = getSuperClass().getPath();
		path.add(this);
		return path;
	}

	@Override
	public List<Path> getPaths() {
		List<Path> paths = new ArrayList<Path>();
		paths.add(getPath());
		for (MIInterface i : getInterfaces())
			paths.addAll(i.getPaths());
		return paths;
	}

	public List<MIMethodTable> getMethodTables() {
		List<MIMethodTable> tables = new ArrayList<MIMethodTable>();
		for (Path p : getPaths())
			tables.add(new MIMethodTable(this, tables.size(), p));
		return tables;
	}

	@Override
	public MIMethod getImplementation(MIMethodSignature sig) {
		MIMethod body = impl.get(sig);
		if (body != null)
			return body;
		if (getSuperClass() != null)
			return getSuperClass().getImplementation(sig);
		return null;
	}

	public boolean hasAbstractMethods() {
		if (declaresAbstractMethods(this))
			return true;
		for (MIType t : getSupertypes())
			if (declaresAbstractMethods(t))
				return true;
		return false;
	}

	private boolean declaresAbstractMethods(MIType t) {
		for (MIMethodSignature sig : t.getMethodSignatures())
			if (getImplementation(sig) == null)
				return true;
		return false;
	}
}
