package de.unibw.i2.mi.oo;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * The abstract super-class of interfaces ({@link MIInterface}) and classes ({@link MIClass}).
 */
public abstract class MIType {
	private final String name;
	private final List<MIMethodSignature> methods = new ArrayList<MIMethodSignature>();

	protected MIType(String name) {
		super();
		this.name = name;
	}

	public String getName() {
		return name;
	}

	public List<MIMethodSignature> getMethodSignatures() {
		return methods;
	}

	/**
	 * Adds the given method signature to this type. Use this method for adding
	 * an abstract method to this type.
	 * 
	 * @see MIClass#addMethod(MIMethod)
	 */
	public void addMethod(MIMethodSignature method) {
		if (methods.contains(method))
			throw new IllegalStateException("adding " + method + " twice to "
					+ this);
		methods.add(method);
	}

	/**
	 * Returns the implementation of the given signature in this type.
	 */
	public abstract MIMethod getImplementation(MIMethodSignature sig);

	/**
	 * Returns the ordered list of all inheritance paths of this type.
	 */
	public abstract List<Path> getPaths();

	/**
	 * Returns the ordered list of direct supertypes of this type.
	 */
	public abstract List<? extends MIType> getDirectSupertypes();

	/**
	 * Returns the collection of all direct or indirect supertypes of this type.
	 * The supertypes are always returned n the same order.
	 */
	public List<MIType> getSupertypes() {
		Set<MIType> supertypes = new LinkedHashSet<MIType>();
		collectSupertypes(supertypes);
		return new ArrayList<MIType>(supertypes);
	}

	/**
	 * Adds all supertypes of this type to the passed set in a deterministic
	 * order.
	 * 
	 * @param supertypes
	 *            in-out parameter: all supertypes are added to this set.
	 */
	private void collectSupertypes(Set<MIType> supertypes) {
		for (MIType t : getDirectSupertypes())
			if (supertypes.add(t))
				t.collectSupertypes(supertypes);
	}

	/**
	 * Returns the displacement for the object pointer of this type when cast to
	 * the given type. Note that displacement values 0, 1, 2, ... are returned.
	 * When adjusting the address of an object pointer, the returned value has
	 * to be multiplied by the number of bytes necessary for storing an address,
	 * e.g., 4 in a 32bit architecture.
	 * 
	 * @return -1 if the given type is not a supertype of or equal to this type.
	 */
	public int upcastDiff(MIType to) {
		int diff = 0;
		for (Path p : getPaths()) {
			if (p.getPath().contains(to))
				return diff;
			diff++;
		}
		return -1;
	}
}
