package de.unibw.i2.mi.oo.parser;

import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.unibw.i2.mi.oo.MIClass;
import de.unibw.i2.mi.oo.MIInterface;
import de.unibw.i2.mi.oo.MIMethod;
import de.unibw.i2.mi.oo.MIMethodSignature;
import de.unibw.i2.mi.oo.MIMethodTable;
import de.unibw.i2.mi.oo.MIType;

public class OO {
	private static final String USAGE = "args: [-v] filename";
	private final List<AClass> types;
	Map<String, MIType> symbols = new HashMap<String, MIType>();
	private int errors;

	private OO(List<AClass> types) {
		super();
		this.types = types;
	}

	/**
	 * args: [-v] filename
	 */
	public static void main(String[] args) {
		String filename;
		boolean verbose = false;
		switch (args.length) {
		case 1:
			filename = args[0];
			break;
		case 2:
			if ("-v".equals(args[0]))
				verbose = true;
			else {
				System.err.println(USAGE);
				return;
			}
			filename = args[1];
			break;
		default:
			System.err.println(USAGE);
			return;
		}

		List<AClass> types;
		try {
			FileReader reader = new FileReader(filename);
			MiniJava parser = new MiniJava(reader);
			types = MiniJava.Program();
		} catch (Throwable ioEx) {
			System.err.println(ioEx.getMessage());
			return;
		}

		if (verbose)
			for (AClass t : types) {
				System.out.println(t);
				System.out.println();
			}

		OO prog = new OO(types);
		prog.translate();
	}

	private void translate() {
		for (AClass t : types) {
			if (symbols.get(t.getName()) != null)
				error("Name " + t.getName() + " used twice.");
			if (t.isInterf())
				makeInterface(t);
			else
				makeClass(t);
		}
		if (errors > 0)
			System.err.println("Aborting due to errors.");
		else
			output();
	}

	private void output() {
		List<MIType> ordered = new ArrayList<MIType>();
		for (AClass t : types)
			if (!t.isInterf()) {
				MIClass cl = (MIClass) symbols.get(t.getName());
				ordered.add(cl);
				methodTables(cl);
			}
		for (AClass t : types)
			if (t.isInterf())
				ordered.add(symbols.get(t.getName()));
		upcastTable(ordered);
	}

	private void upcastTable(List<MIType> ordered) {
		// determine longest name
		int max = 0;
		for (MIType t : ordered)
			max = Math.max(max, t.getName().length());
		// set printf format strings
		String sFormat1 = "%-" + max + "s ||";
		String sFormat2 = " %" + max + "s |";
		String dFormat = " %" + max + "d |";

		System.out.println("Upcast table:");
		// header row
		System.out.printf(sFormat1, " ");
		for (MIType t : ordered)
			System.out.printf(sFormat2, t.getName());
		System.out.println();

		line('=', ordered, max);

		for (MIType from : ordered) {
			// row
			System.out.printf(sFormat1, from.getName());
			for (MIType to : ordered) {
				int diff = from.upcastDiff(to);
				if (diff < 0)
					System.out.printf(sFormat2, "-");
				else
					System.out.printf(dFormat, 4 * diff);
			}
			System.out.println();
			line('-', ordered, max);
		}
	}

	private void line(char c, List<MIType> ordered, int max) {
		for (int i = 0; i <= max; i++)
			System.out.print(c);
		System.out.print("++");
		for (MIType t : ordered) {
			for (int i = 0; i < max+2; i++)
				System.out.print(c);
			System.out.print('+');
		}
		System.out.println();
	}

	private void methodTables(MIClass aClass) {
		System.out.println("Class " + aClass.getName());
		if (aClass.isAbstract())
			System.out.println("  abstract class ==> no method tables");
		else
			for (MIMethodTable table : aClass.getMethodTables()) {
				if (table.isClassMethodTable())
					System.out.println("  class method table:");
				else {
					System.out.println("  interface method table for "
							+ table.getPath());
					System.out.println("    delta = " + 4 * table.getDelta());
				}
				for (MIMethod m : table.getMethods()) {
					System.out.print("    ");
					System.out.println(m.shortString());
				}
			}
		System.out.println();
	}

	private List<MIInterface> makeInterfaceList(AClass t) {
		List<MIInterface> interfaces = new ArrayList<MIInterface>();
		for (String name : t.getInterfaces()) {
			MIType iType = symbols.get(name);
			if (iType == null)
				error("Unknown interface " + iType);
			else if (!(iType instanceof MIInterface))
				error("not an interface: " + name);
			else
				interfaces.add((MIInterface) iType);
		}
		return interfaces;
	}

	private void makeClass(AClass t) {
		List<MIInterface> interfaces = makeInterfaceList(t);
		MIClass superClass = null;
		if (t.getSuperClass() != null) {
			MIType sType = symbols.get(t.getSuperClass());
			if (sType == null)
				error("Unknown class " + t.getSuperClass());
			else if (!(sType instanceof MIClass))
				error("not a class: " + t.getSuperClass());
			else
				superClass = (MIClass) sType;
		}
		MIClass dClass = new MIClass(t.getName(), t.isAbstract(), superClass,
				interfaces);
		symbols.put(t.getName(), dClass);
		addMethods(dClass, t);
		if (!t.isAbstract() && dClass.hasAbstractMethods())
			error(t.getName() + " must be declared abstract.");
	}

	private void makeInterface(AClass t) {
		List<MIInterface> interfaces = makeInterfaceList(t);
		MIInterface dInterf = new MIInterface(t.getName(), interfaces);
		symbols.put(t.getName(), dInterf);
		addMethods(dInterf, t);
	}

	private void addMethods(MIInterface dInterf, AClass t) {
		for (AMethod m : t.getMethods()) {
			MIMethodSignature sig = new MIMethodSignature(m.getName());
			dInterf.addMethod(sig);
		}
	}

	private void addMethods(MIClass dClass, AClass t) {
		for (AMethod m : t.getMethods()) {
			MIMethodSignature sig = new MIMethodSignature(m.getName());
			if (m.isAbstractMethod()) {
				if (!dClass.isAbstract())
					error("Class "
							+ dClass.getName()
							+ " must be declared abstract because of abstract method "
							+ m.getName());
				else
					dClass.addMethod(sig);
			} else {
				MIMethod meth = new MIMethod(dClass, sig);
				dClass.addMethod(meth);
			}
		}
	}

	private void error(String string) {
		errors++;
		System.err.println(string);
	}

}
